// Eltek Monitor / Control based on code from https://github.com/the6p4c/Flatpack2
// Where a breakout board for the connections can also be found
// Protocol is here https://github.com/The6P4C/Flatpack2/blob/master/Protocol.md
// Modified and adapted by G0MJW

#include <mcp_can.h>
#include <U8g2lib.h>

#include <SPI.h>    // SPI bus
#include <Wire.h>   // Wire is required for i2c LCD Display
#include "eltek.h"  // Project specific defines

// LCD Parameters U8G2_ST7920_128X64_F_8080 = 1024 bytes - Full buffer - use u8g2.sendBuffer();
//                U8G2_ST7920_128X64_1_8080 =  128 bytes - Page buffer - use do-while loop
//                U8G2_ST7920_128X64_2_8080 =  256 bytes - Page buffer - use do-while loop

U8G2_ST7920_128X64_F_8080 u8g2(U8G2_R0, LCD_D0, LCD_D1, LCD_D2, LCD_D3, LCD_D4, LCD_D5, LCD_D6, LCD_D7, LCD_EN, /*cs=*/U8X8_PIN_NONE, LCD_RS, LCD_RST);  // Remember to set R/W to 0

// CAN BUS Parameters
MCP_CAN CAN(CAN_CS_PIN);

// Globals
bool serialNumberReceived = false;  // PSU detected and serial number found
uint8_t serialNumber[6];            // The serial number as an array of nibbles
unsigned long lastLogInTime = 0;    // Timer for loging in
int intakeTemperature = 0;          // Air temperature
float current = 0;                  // Output current
float outputVoltage = 0;            // Output Voltage
int inputVoltage = 0;               // Mains input voltage
int outputTemperature = 0;          // PSU temperature
int sigLoss = TIMEOUT;              // if it goes round the loop more times than this link lost
int lineht;                         // For LCD writing
bool initialised = false;           // Initialise CAN bus
uint8_t status, last_status;        // Status tracking
uint8_t alarms0, alarms1;           // Alarms and warnings

/*-------------------------------------------------------------------------*/
void setup() {
  status = STATE_INIT;
  last_status = status;
  Serial.begin(9600);    
  pinMode(CAN_CS_PIN, OUTPUT);
  pinMode(CAN_INT_PIN, INPUT);

  // Initialise and print splash screen on LCD
  lcdStartup();
  if (CAN.begin(MCP_ANY, CAN_125KBPS, MCP_8MHZ) == CAN_OK)
    Serial.println(F("MCP2515 initialized successfully."));
  else {
    Serial.println(F("Failed to initialize MCP2515. Halting."));
    last_status = status;
    status = STATE_FAIL;
    showFail();
    while (1)
      ;  // Loop forever
  }
  CAN.setMode(MCP_NORMAL);
  lastLogInTime = millis();
}
/*-------------------------------------------------------------------------*/
void printMessage(uint32_t rxID, uint8_t len, uint8_t rxBuf[]) {
  char output[256];

  snprintf(output, 256, "ID: 0x%.8lX Length: %1d Data:", rxID, len);
  Serial.print(output);
  for (int i = 0; i < len; ++i) {
    snprintf(output, 256, " 0x%.2X", rxBuf[i]);
    Serial.print(output);
  }
  Serial.println();
}
/*-------------------------------------------------------------------------*/
void logIn() {
#ifdef DEBUG
  Serial.println(F("--------"));
  Serial.println(F("Logging in."));
  Serial.println(F("--------"));
#endif

  uint8_t txBuf[8] = { 0 };
  for (int i = 0; i < 6; ++i) {
    txBuf[i] = serialNumber[i];
  }
  CAN.sendMsgBuf(0x05004804, 1, 8, txBuf);
}
/*-------------------------------------------------------------------------*/
void lcdStartup() {
  u8g2.begin(); 
  u8g2.firstPage();
  do {
    drawLogo();
  } while (u8g2.nextPage());
  delay(5000);
  u8g2.clear();  //clear display
  do {
    u8g2.drawFrame(30, 20, 65, 16);
    u8g2.setFont(u8g2_font_6x13_tr);  // Set the font
    u8g2.drawStr(34, 32, "Searching");
  } while (u8g2.nextPage());
}
/*-------------------------------------------------------------------------*/
void lcdInit() {
  // Shows the info Screen
  char output[3];
  u8g2.clear();  //clear display
  do {
    u8g2.drawFrame(0, 0, 127, 64);
    u8g2.setFont(u8g2_font_6x13_tr);  // Set the font
    u8g2.drawStr(6, 12, "G0MJW Eltek Control");
    u8g2.drawLine(1, 14, 127, 14);
    u8g2.setFont(u8g2_font_5x8_tr);  // Set the font
    u8g2.drawStr(8, 30, "Version 1 - July 2023");
    u8g2.drawStr(8, 40, "CAN BUS MCP2115");
    u8g2.drawStr(8, 50, "PSU Found");
    u8g2.drawStr(8, 60, "S/N ");
    for (int i = 0; i < 6; ++i) {
      snprintf(output, 3, "%.2X", serialNumber[i]);
      u8g2.drawStr(26 + i * 12, 60, output);
    }
  } while (u8g2.nextPage());
  delay(5000);  // Just for testing 
}
/*-------------------------------------------------------------------------*/
void drawLogo(void) {
  // Displays a splash screen 
  u8g2.clear();
  do {
    u8g2.drawFrame(0, 0, 127, 49);
    //u8g2.drawFrame(1, 1, 126, 49);
    u8g2.drawLine(1, 1, 126, 1);
    u8g2.drawLine(1, 1, 1, 48);
    u8g2.drawCircle(30, 24, 22, U8G2_DRAW_ALL);
    u8g2.drawCircle(30, 24, 16, U8G2_DRAW_ALL);
    u8g2.drawCircle(30, 24, 7, U8G2_DRAW_ALL);
    u8g2.drawLine(10, 14, 40, 44);
    u8g2.drawLine(50, 14, 20, 44);
    u8g2.drawLine(50, 34, 20,  4);
    u8g2.drawLine(10, 34, 40,  4);
    u8g2.setFont(u8g2_font_5x7_tr);  // Set the font
    u8g2.drawStr(60, 10, "FLATPACK2 HE");
    u8g2.drawStr(70, 18, "48V/2000W");
    u8g2.drawDisc(60, 36, 2, U8G2_DRAW_ALL);
    u8g2.drawDisc(60, 30, 2, U8G2_DRAW_ALL);
    u8g2.drawDisc(60, 24, 2, U8G2_DRAW_ALL);
    u8g2.setFont(u8g2_font_crox2c_tr);
    u8g2.drawStr(6, 61, "G0MJW Designs");    
    } while (u8g2.nextPage());
}
/*-------------------------------------------------------------------------*/
void showFail() {
  if(status != last_status)
    {  
    char output[3];
    do {
      u8g2.drawFrame(0, 0, 127, 63);
     u8g2.setFont(u8g2_font_6x13_tr);  // Set the font
     u8g2.drawStr(4, 30, "Initialsation Failed");
     u8g2.drawStr(64, 50, "Halted");
    } while (u8g2.nextPage());
    }
   last_status = status;
}
/*-------------------------------------------------------------------------*/
void lcdFrame() {
  // Sets up the LCD frame
  u8g2.clear();  
  do {
    u8g2.drawFrame(0, 0, 127, 64);
    u8g2.setFont(u8g2_font_6x13_tr);  // Set the font
    u8g2.drawStr(6, 12, "Eltek PSU Controller");
    u8g2.drawLine(1, 14, 127, 14);
    u8g2.drawLine(1, 46, 127, 46);
    u8g2.drawLine(76, 14, 76, 63);
  } while (u8g2.nextPage());  
}
/*-------------------------------------------------------------------------*/
void lcdCommFail() {
  if(status != last_status)
    {    
    // Screen for communications failure
    u8g2.clear();    
    do {
      u8g2.drawFrame(30, 20, 65, 16);
      u8g2.setFont(u8g2_font_6x13_tr);  // Set the font
      u8g2.drawStr(34, 32, "Link Lost!");
      } while (u8g2.nextPage());
    }
    last_status = status;
}
/*-------------------------------------------------------------------------*/
void processLogInRequest(uint32_t rxID, uint8_t len, uint8_t rxBuf[]) {
  Serial.println(F("--------"));
  Serial.print(F("Found power supply "));

  char output[3];

  for (int i = 0; i < 6; ++i) {
    serialNumber[i] = rxBuf[i + 1];

    snprintf(output, 3, "%.2X", serialNumber[i]);
    Serial.print(output);
  }
  last_status = status;
  status = STATE_DETECT;
  if (!serialNumberReceived)  // Only do this once
    lcdInit();
  serialNumberReceived = true;    
  Serial.println();
  Serial.println(F("--------"));
}
/*-------------------------------------------------------------------------*/
void processStatusMessage(uint32_t rxID, uint8_t len, uint8_t rxBuf[]) {
  // Decodes the status message and sets variables
  intakeTemperature = rxBuf[0];
  current = 0.1f * (rxBuf[1] | (rxBuf[2] << 8));
  outputVoltage = 0.01f * (rxBuf[3] | (rxBuf[4] << 8));
  inputVoltage = rxBuf[5] | (rxBuf[6] << 8);
  outputTemperature = rxBuf[7];
  bool hasWarning = rxID == 0x05014008;
  bool hasAlarm = rxID == 0x0501400C;
  last_status = status;
  if (hasWarning)     
    Serial.println("WARNING");        
   else if (hasAlarm)     
    Serial.println("ALARM");     
          
  if (hasWarning || hasAlarm)  // Say what the issue is
    {
    uint8_t txBuf[3] = { 0x08, hasWarning ? 0x04 : 0x08, 0x00 };
    CAN.sendMsgBuf(0x0501BFFC, 1, 3, txBuf);
    status = STATE_ALARM;
    }
    else
      status = STATE_RUN;
}
/*-------------------------------------------------------------------------*/
void processWarningOrAlarmMessage(uint32_t rxID, uint8_t len, uint8_t rxBuf[]) {
  bool isWarning = rxBuf[1] == 0x04;
  alarms0 = rxBuf[3];
  alarms1 = rxBuf[4];

  if (isWarning)    
    Serial.print(F("Warnings:"));
    
  else    
    Serial.print(F("Alarms:"));    

  for (uint8_t i = 0; i < 8; ++i) {
    if (alarms0 & (1 << i)) {
      Serial.print(" ");
      Serial.print(alarms0Strings[i]);
    }
    if (alarms1 & (1 << i)) {
      Serial.print(" ");
      Serial.print(alarms1Strings[i]);
    }
  }
}
/*-------------------------------------------------------------------------*/
void lcdShowAlarms()
{
  // Shows the alarms
if(status != last_status)
  {
  u8g2.clear();    
      do {        
        u8g2.setFont(u8g2_font_4x6_tr);  // Set the font
        u8g2.drawStr(0, 6, "Warnings");
        u8g2.drawStr(64, 6, "Alarms");
        for (uint8_t i = 0; i < 8; ++i) 
          {
          if (alarms0 & (1 << i))             
            u8g2.drawStr(0, 16 + i * 6, alarms0Strings[i]);            
          if (alarms1 & (1 << i)) 
            u8g2.drawStr(64, 16 + i * 6, alarms1Strings[i]);
          }
      } while (u8g2.nextPage());
  }
  last_status = status;
}
/*-------------------------------------------------------------------------*/
void setVoltage(uint16_t Voltage) {
  // Sets voltage, note this is in millivolts
  logIn();
  delay(100);
  uint8_t voltageSetTxBuf[5] = { 0x29, 0x15, 0x00, Voltage & 0xFF, (Voltage >> 8) & 0xFF };
  CAN.sendMsgBuf(0x05019C00, 1, 5, voltageSetTxBuf);
}
/*-------------------------------------------------------------------------*/
void printStatus(uint32_t rxID) {
  // Prints a status message on the serial port
  Serial.print(F("Status: Air In "));
  Serial.print(intakeTemperature);
  Serial.print(F("C Out "));
  Serial.print(outputTemperature);
  Serial.print(F("C | Current "));
  Serial.print(current);
  Serial.print(F("A | Voltage In "));
  Serial.print(inputVoltage);
  Serial.print(F("V | Out "));
  Serial.print(outputVoltage);
  Serial.println("V");

  if (rxID == 0x05014010)
    Serial.println(F("Currently in walk in (voltage ramping up)"));
}
/*-------------------------------------------------------------------------*/
void lcdShowData() {
  char text[21];
    // Frame for data - only update if necessary, i.e. when running and only once
  if (status != last_status)  
    lcdFrame();  
  
  int VV = (int)outputVoltage;
  int vv = (int)(outputVoltage * 10) % 10;
  int AA = (int)current;
  int aa = (int)(current * 10) % 10;
  int power = outputVoltage * current;
  do { 
    u8g2.setFont(u8g2_font_crox5hb_tr);  // Set the large font
    sprintf(text, "%02d.%01dA", AA, aa);
    u8g2.drawStr(6, 40, text);    
    u8g2.setFont(u8g2_font_crox2c_tr);  // Set the large font
    sprintf(text, "%02d.%01dV %04dW", VV, vv);
    u8g2.drawStr(78, 36, text);       
    sprintf(text, "%04dW", power);
    u8g2.drawStr(16, 60, text);   
    u8g2.setFont(u8g2_font_5x8_tr);  // Set the medium font    
    sprintf(text, "%03dV %02dC", inputVoltage,  outputTemperature);
    u8g2.drawStr(82, 58, text); 
  } while (u8g2.nextPage());
}
/*-------------------------------------------------------------------------*/
void loop() {
  uint32_t rxID;
  uint8_t len = 0;
  uint8_t rxBuf[8]; 

  if (millis() - lastLogInTime > 1000) {
    last_status = status;
    sigLoss--;
    lastLogInTime = millis();
    if (serialNumberReceived)
      logIn();

    if (status != STATE_COMS_LOST)  // print status on serial line
      printStatus(rxID);
  }
  if (sigLoss <= 0) 
    {
    // If here we have lost comms
    if (status != STATE_COMS_LOST) 
      {
      sigLoss = 0;      
      status = STATE_COMS_LOST;
      Serial.println(F("Comms Failure"));
      lcdCommFail() ;       
      }
  }
  // If something in buffer goes INT low
  if (!digitalRead(CAN_INT_PIN)) {
    CAN.readMsgBuf((unsigned long *)&rxID, &len, rxBuf);

    // Limit ID to lowest 29 bits (extended CAN)
    rxID &= 0x1FFFFFFF;
#ifdef DEBUG
    printMessage(rxID, len, rxBuf);
#endif
    if (!serialNumberReceived && (rxID & 0xFFFF0000) == 0x05000000)  // Log in first time
      {
      processLogInRequest(rxID, len, rxBuf);
      sigLoss = TIMEOUT;  // reset timeout
      }
    else if ((rxID & 0xFFFFFF00) == 0x05014000)  // Status data
      {
      processStatusMessage(rxID, len, rxBuf);
      if (status == STATE_RUN && serialNumberReceived)        
        lcdShowData();
      sigLoss = TIMEOUT;  // reset timeout  
      }
     else if (rxID == 0x0501BFFC)  // Alarms & warmings
        {
        processWarningOrAlarmMessage(rxID, len, rxBuf);
        if(status == STATE_ALARM)
          lcdShowAlarms();   
        sigLoss = TIMEOUT;  // reset timeout    
        }
    else if ((rxID & 0xFFFF0000) == 0x05000000)  // Log in success
      {
      sigLoss = TIMEOUT;  // reset timeout
      last_status = status;
      status == STATE_RUN;
      }
  }
}
/*-------------------------------------------------------------------------*/
